/* sbrk-mutex.c
 *
 * Create NUM_THREADS to walk through an array of malloc'd pthread mutexes.
 * Each thread holds up to NUM_CONCURRENT locks at a time.
 *
 * build with: gcc sbrk-mutex.c -lrt -o sbrk-mutex
 *
 * 2006-02-28  Initial version by Darren Hart
 * 1006-03-01  Changed mutexes to PTHREAD_MUTEX_ROBUST_NP type - Sripathi Kodi
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <time.h>
#include <pthread.h>
#include <sched.h>
#include <errno.h>
#include <unistd.h>

#define NUM_MUTEXES 5000
#define NUM_THREADS 50
#define NUM_CONCURRENT_LOCKS 50
#define DELAY 1000 /* how long to sleep in the worker thread in us */

static pthread_mutex_t *mutexes[NUM_MUTEXES];
static pthread_t threads[NUM_THREADS];


void *worker_thread(void *arg)
{
	int i;

	for (i = 0; i < NUM_MUTEXES + NUM_CONCURRENT_LOCKS; i++) {
		/* release prior lock */
		if (i >= NUM_CONCURRENT_LOCKS) {
			pthread_mutex_unlock(mutexes[i-NUM_CONCURRENT_LOCKS]);
		}
		/* grab a new lock */
		if (i < NUM_MUTEXES) {
			pthread_mutex_lock(mutexes[i]);
		}

		usleep(DELAY);
		printf("thread %ld @ %d\n", (long)arg, i);
	}
	return NULL;
}

int main(int argc, char* argv[])
{
	int m, ret, robust;
	intptr_t t;
	pthread_mutexattr_t mutexattr;

        if (pthread_mutexattr_init(&mutexattr) != 0) {
                printf("Failed to init mutexattr\n");
        }
        if (pthread_mutexattr_setrobust_np(&mutexattr, PTHREAD_MUTEX_ROBUST_NP) != 0) {
                printf("Can't set mutexattr robust\n");
        }
        if (pthread_mutexattr_getrobust_np(&mutexattr, &robust) != 0) {
                printf("Can't get mutexattr robust\n");
        } else {
                printf("robust in mutexattr is %d\n", robust);
        }


	/* malloc and initialize the mutexes */
	printf("allocating and initializing %d mutexes\n", NUM_MUTEXES);
	for (m = 0; m < NUM_MUTEXES; m++) {
		if (!(mutexes[m] = malloc(sizeof(pthread_mutex_t)))) {
			perror("malloc failed\n");
		}
		if ((ret = pthread_mutex_init(mutexes[m], &mutexattr))) {
			perror("pthread_mutex_init() failed\n");
		}
	}
	printf("mutexes allocated and initialized successfully\n");

	/* start children threads to walk the array, grabbing the locks */
	for (t = 0; t < NUM_THREADS; t++) {
		pthread_attr_t attr;
		struct sched_param param;

		param.sched_priority = sched_get_priority_min(SCHED_FIFO);
		pthread_attr_init(&attr);
		pthread_attr_setinheritsched (&attr, PTHREAD_EXPLICIT_SCHED);
		pthread_attr_setschedparam(&attr, &param);
		pthread_attr_setschedpolicy(&attr, SCHED_FIFO);

		if (pthread_create(&threads[t], &attr, worker_thread, (void*)t)) {
			perror("pthread_create failed");
		}

		pthread_attr_destroy(&attr);
	}

	/* wait for the children to complete */
	printf("joining threads");
	for (t = 0; t < NUM_THREADS; t++) {
		pthread_join(threads[t], 0);
	}

	/* destroy all the mutexes */
	for (m = 0; m < NUM_MUTEXES; m++) {
		if (mutexes[m]) {
			if ((ret = pthread_mutex_destroy(mutexes[m])))
				perror("pthread_mutex_destroy() failed\n");
			free(mutexes[m]);
		}
	}
	return 0;
}
